#include <iomanip.h>
#include <iostream.h>
#include <limits.h>
#include <pthread.h>
#include <regex.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "check_version.h"
#include "Server.H"

int Compile_regex (regex_t *preg, char *pattern, char icase, char ext)
{
  int err = regcomp (preg, pattern,
		     (icase ? REG_ICASE : 0) | (ext ? REG_EXTENDED : 0));
  
  if (err)
    {
      char buf[100];
      regerror (err, preg, buf, 100);
      cerr << buf << endl;
      return 1;
    }
  
  return 0;
}

int GetConf (Server &kom, char *name, char f)
{
  Token *token;
  
  kom.Command (Server_LookupZName, name, 1, f);
  token = kom.Read ();
  
  switch (token->Array (0)->Number ())
    {
    case 0:
      cerr << "No match for " << name << endl;
      return -1;

    case 1:
      int n;
      n = token->Array (1)->Array (2)->Number ();
      delete token;
      return n;
      
    default:
      int cntr = 0;
      cerr << "Multiple matches for " << name << endl;
      while (cntr < token->Array (0)->Number ())
	{
	  cerr << token->Array (1)->Array (cntr * 3)->String () << endl;
	  cntr++;
	}
      return -1;
    }
}

Token *GetTextdata (Server &kom, int conf)
{
  kom.Command (Server_GetConfStat, conf);
  Token *stat = kom.Read ();
  
  kom.Command (Server_LocalToGlobal, conf, 1, stat->Array (30)->Number ());
  Token *txts = kom.Read ();
  
  delete stat;
  return txts;
}

void Help (char help)
{
  cerr << "komgrep version " << VERSION << endl
       << "Usage: komgrep USERNAME PATTERN CONFERENCE" << endl;
  if (help)
    {
      cerr << "Search for PATTERN in CONFERENCE." << endl
	   << "Example: komgrep 'Eskil Block' 'experi.*brytas' 'h m p'" << endl
	   << "  -E    Use extended regular expression syntax" << endl
	   << "  -i    ignore case" << endl
	   << "  -k    komserver" << endl
	   << "  -n    search throgh the last n article (*)" << endl
	   << "  -p    portnumber" << endl
	   << "  -r    reverse search order (*)" << endl
	   << endl << "        (*) not implemented yet..." << endl;
    }
  else
    {
      cerr << "-h for more help." << endl;
    }
}

void Login (Server &kom, int pers)
{
  Token *token;
  char *pass;
  pass = getpass ("Password: ");
  if (pass == NULL || strlen (pass) == 0)
    {
      cerr << "No password. Quitting." << endl;
      exit (EXIT_SUCCESS);
    }
  kom.Command (Server_Login, pers, pass, 1);
  token = kom.Read ();
  delete token;
}

void Search (Server &kom, int pers, int conf, regex_t *preg)
{
  int tc = 1;
  int cnt;
  int dns;
  int max;
  int txtnr;
  regmatch_t pmatch[1];
  
  Token *txts;
  do
    {
      kom.Command (Server_LocalToGlobal, conf, tc, 255);
      txts = kom.Read ();
      cnt = txts->Array (2)->Number ();
      dns = txts->Array (3)->Number ();
      max = txts->Array ( dns ? 5 : 4)->Number ();

      int cntr = 0;
      while (cntr < max)
	{
	  txtnr = txts->Array (dns ? 6 : 5)->
	    Array (cntr * (dns ? 1 : 2) + (dns ? 0 : 1))->Number ();
	  if (txtnr)
	    {
	      cerr << setw (8) << txtnr << "\r";
	      kom.Command (Server_GetText, txtnr, 0, INT_MAX);
	      Token *txt = kom.Read ();
	      char *p;
	      p = txt->Array (0)->String ();
	      while (*p)
		{
		  if (*p == '\n') *p = ' ';
		  p++;
		}
	      char buf[71];
	      if (!regexec (preg, txt->Array (0)->String (), 1, pmatch, 0))
		{
		  int len = pmatch->rm_eo - pmatch->rm_so;
		  if (len > 70) len = 70;
		  p = txt->Array (0)->String () + pmatch->rm_so
		    - (70 - len) / 2;
		  if (strlen (p) < 70) p -= 70 - strlen (p);
		  if (p < txt->Array (0)->String ())
		    p = txt->Array (0)->String ();
		  strncpy (buf, p, 70);
		  buf[70] = '\0';
		  cout << setw (8) << txtnr;
		  cout << ":" << buf << endl;
		}
	    }
	  cntr++;
	}
      tc = txts->Array (1)->Number ();
      delete txts;
    }
  while (cnt);
}

int main (int argc, char **argv)
{
  char *server = NULL;
  char ext = 0;
  char help = 0;
  char icase = 0;
  unsigned short port = 4894;
  regex_t preg;
  
  while (1)
    {
      int c;
      c = getopt (argc, argv, "Ehik:p:");
      if (c < 0) break;
      switch (c)
	{
	case 'E':
	  ext = 1;
	  break;
	  
	case 'h':
	  help = 1;
	  break;
	  
	case 'i':
	  icase = 1;
	  break;
	  
	case 'k':
	  server = new char[strlen (optarg) + 1];
	  strcpy (server, optarg);
	  break;
	  
	case 'p':
	  port = atoi (optarg);
	  if (!port)
	    {
	      cerr << "Illegal port \"" << optarg << "\"." << endl;
	      return EXIT_FAILURE;
	    }
	  break;
	  
	default:
	  cerr << "Flag From Hell<TM>! I've never heard of -"
	       << (char) c << "." << endl;
	  return EXIT_FAILURE;
	}
    }
  
  if (argc - optind < 3)
    {
      Help (help);
      return EXIT_FAILURE;
    }
  
  Server kom (server ? server : "kom.lysator.liu.se", port);
  
  char *pers_s = argv[optind];
  char *pattern = argv[optind + 1];
  char *conf_s = argv[optind + 2];
  
  if (Compile_regex (&preg, pattern, icase, ext)) return 1;
  
  kom.Command (Server_InitProtocol, pers_s, NULL);
  
  int pers = GetConf (kom, pers_s, 0);
  if (pers < 0) return EXIT_FAILURE;
  
  Login (kom, pers);
  
  int conf = GetConf (kom, conf_s, 1);
  if (conf < 0) return EXIT_FAILURE;
  
  pthread_t cv;
  pthread_create (&cv, NULL, check_version, "komgrep");
  
  Search (kom, pers, conf, &preg);
  
  if (latest_version[0])
    cerr << "There's an updated version of komgrep available ("
	 << latest_version << ")." << endl;
  return EXIT_SUCCESS;
}
